@@ -65,7 +65,7 @@ int get_notify_fd (libcrun_context_t *context, libcrun_container_t *container, i
                    libcrun_error_t *err);
 int libcrun_set_mounts (struct container_entrypoint_s *args, libcrun_container_t *container, const char *rootfs,
                         set_mounts_cb_t cb, void *cb_data, libcrun_error_t *err);
-int libcrun_init_caps (libcrun_error_t *err);
+int libcrun_init_caps (libcrun_container_t *container, libcrun_error_t *err);
 int libcrun_do_pivot_root (libcrun_container_t *container, bool no_pivot, const char *rootfs, libcrun_error_t *err);
 int libcrun_reopen_dev_null (libcrun_error_t *err);
 int libcrun_set_usernamespace (libcrun_container_t *container, pid_t pid, libcrun_error_t *err);
@@ -73,8 +73,8 @@ int libcrun_set_caps (runtime_spec_schema_config_schema_process_capabilities *ca
                       int no_new_privileges, libcrun_error_t *err);
 int libcrun_set_rlimits (runtime_spec_schema_config_schema_process_rlimits_element **rlimits, size_t len,
                          libcrun_error_t *err);
-int libcrun_set_selinux_label (runtime_spec_schema_config_schema_process *proc, bool now, libcrun_error_t *err);
-int libcrun_set_apparmor_profile (runtime_spec_schema_config_schema_process *proc, bool now, libcrun_error_t *err);
+int libcrun_set_selinux_label (libcrun_container_t *container, runtime_spec_schema_config_schema_process *proc, bool now, libcrun_error_t *err);
+int libcrun_set_apparmor_profile (libcrun_container_t *container, runtime_spec_schema_config_schema_process *proc, bool now, libcrun_error_t *err);
 int libcrun_set_hostname (libcrun_container_t *container, libcrun_error_t *err);
 int libcrun_set_domainname (libcrun_container_t *container, libcrun_error_t *err);
 int libcrun_set_oom (libcrun_container_t *container, libcrun_error_t *err);
@@ -87,7 +87,7 @@ int libcrun_linux_container_update (libcrun_container_status_t *status,
                                     const char *state_root,
                                     runtime_spec_schema_config_linux_resources *resources,
                                     libcrun_error_t *err);
-int libcrun_create_keyring (const char *name, const char *label, libcrun_error_t *err);
+int libcrun_create_keyring (libcrun_container_t *container, const char *name, const char *label, libcrun_error_t *err);
 int libcrun_container_pause_linux (libcrun_container_status_t *status, libcrun_error_t *err);
 int libcrun_container_unpause_linux (libcrun_container_status_t *status, libcrun_error_t *err);
 int libcrun_container_do_bind_mount (libcrun_container_t *container, char *mount_source, char *mount_destination,
@@ -127,26 +127,22 @@ enum
 
 #define LIBCRUN_INTELRDT_CREATE_UPDATE_MOVE (LIBCRUN_INTELRDT_CREATE | LIBCRUN_INTELRDT_UPDATE | LIBCRUN_INTELRDT_MOVE)
 
-static inline bool
-container_has_intelrdt (libcrun_container_t *container)
-{
-  runtime_spec_schema_config_schema *def = NULL;
-
-  def = container->container_def;
-
-  return def != NULL && def->linux != NULL && def->linux->intel_rdt != NULL;
-}
-
-const char *libcrun_get_intelrdt_name (const char *ctr_name, libcrun_container_t *container, bool *explicit);
-
 int libcrun_apply_intelrdt (const char *ctr_name, libcrun_container_t *container, pid_t pid, int actions, libcrun_error_t *err);
 
-int libcrun_destroy_intelrdt (const char *name, libcrun_error_t *err);
+int libcrun_move_network_devices (libcrun_container_t *container, pid_t pid, libcrun_error_t *err);
 
-int libcrun_update_intel_rdt (const char *ctr_name, libcrun_container_t *container, const char *l3_cache_schema, const char *mem_bw_schema, libcrun_error_t *err);
+int libcrun_destroy_intelrdt (const char *container_id, runtime_spec_schema_config_schema *def, libcrun_error_t *err);
+
+int libcrun_update_intel_rdt (const char *ctr_name, libcrun_container_t *container, const char *l3_cache_schema, const char *mem_bw_schema, char *const *schemata, libcrun_error_t *err);
 
 int libcrun_safe_chdir (const char *path, libcrun_error_t *err);
 
-int get_bind_mount (int dirfd, const char *src, bool recursive, bool rdonly, libcrun_error_t *err);
+int get_bind_mount (int dirfd, const char *src, bool recursive, bool rdonly, bool nofollow, libcrun_error_t *err);
+
+bool is_bind_mount (runtime_spec_schema_defs_mount *mnt, bool *recursive, bool *src_nofollow);
+
+int libcrun_make_runtime_mounts (libcrun_container_t *container, libcrun_container_status_t *status, runtime_spec_schema_defs_mount **mounts, size_t len, libcrun_error_t *err);
+
+int libcrun_destroy_runtime_mounts (libcrun_container_t *container, libcrun_container_status_t *status, runtime_spec_schema_defs_mount **mounts, size_t len, libcrun_error_t *err);
 
 #endif
