@@ -50,6 +50,7 @@
 #else
 #  define atomic_long volatile long
 #endif
+#include "syscalls.h"
 
 #ifndef CLOSE_RANGE_CLOEXEC
 #  define CLOSE_RANGE_CLOEXEC (1U << 2)
@@ -66,29 +67,6 @@
 
 #define MAX_READLINKS 32
 
-static int
-syscall_close_range (unsigned int fd, unsigned int max_fd, unsigned int flags)
-{
-  return (int) syscall (__NR_close_range, fd, max_fd, flags);
-}
-
-static int
-syscall_openat2 (int dirfd, const char *path, uint64_t flags, uint64_t mode, uint64_t resolve)
-{
-  struct openat2_open_how
-  {
-    uint64_t flags;
-    uint64_t mode;
-    uint64_t resolve;
-  } how = {
-    .flags = flags,
-    .mode = mode,
-    .resolve = resolve,
-  };
-
-  return (int) syscall (__NR_openat2, dirfd, path, &how, sizeof (how), 0);
-}
-
 int
 crun_path_exists (const char *path, libcrun_error_t *err)
 {
@@ -126,7 +104,7 @@ write_file_at_with_flags (int dirfd, int flags, mode_t mode, const char *name, c
 
   fd = openat (dirfd, name, O_CLOEXEC | O_WRONLY | flags, mode);
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "opening file `%s` for writing", name);
+    return crun_make_error (err, errno, "open `%s` for writing", name);
 
   ret = safe_write (fd, name, data, len, err);
   if (UNLIKELY (ret < 0))
@@ -228,7 +206,7 @@ create_file_if_missing_at (int dirfd, const char *file, mode_t mode, libcrun_err
       if (ret == 0 && S_ISREG (tmp_mode))
         return 0;
 
-      return crun_make_error (err, errno, "creating file `%s`", file);
+      return crun_make_error (err, errno, "create file `%s`", file);
     }
   return 0;
 }
@@ -246,6 +224,7 @@ ensure_directory_internal_at (int dirfd, char *path, size_t len, int mode, libcr
       if (ret == 0 || errno == EEXIST)
         return 0;
 
+      int saved_errno = errno;
       if (parent_created || errno != ENOENT)
         {
           libcrun_error_t tmp_err = NULL;
@@ -257,7 +236,7 @@ ensure_directory_internal_at (int dirfd, char *path, size_t len, int mode, libcr
           if (ret < 0)
             crun_error_release (&tmp_err);
 
-          return crun_make_error (err, errno, "create directory `%s`", path);
+          return crun_make_error (err, saved_errno, "create directory `%s`", path);
         }
 
       while (it > path && *it != '/')
@@ -298,6 +277,25 @@ crun_ensure_directory_at (int dirfd, const char *path, int mode, bool nofollow,
   return 0;
 }
 
+static int
+check_fd_is_path (const char *path, int fd, const char *fdname, libcrun_error_t *err)
+{
+  proc_fd_path_t fdpath;
+  size_t path_len = strlen (path);
+  char link[PATH_MAX];
+  int ret;
+
+  get_proc_self_fd_path (fdpath, fd);
+  ret = TEMP_FAILURE_RETRY (readlink (fdpath, link, sizeof (link)));
+  if (UNLIKELY (ret < 0))
+    return crun_make_error (err, errno, "readlink `%s`", fdname);
+
+  if (((size_t) ret) != path_len || memcmp (link, path, path_len))
+    return crun_make_error (err, 0, "target `%s` does not point to the directory `%s`", fdname, path);
+
+  return 0;
+}
+
 static int
 check_fd_under_path (const char *rootfs, size_t rootfslen, int fd, const char *fdname, libcrun_error_t *err)
 {
@@ -377,6 +375,23 @@ safe_openat (int dirfd, const char *rootfs, const char *path, int flags, int mod
   static bool openat2_supported = true;
   int ret;
 
+  if (is_empty_string (path))
+    {
+      cleanup_close int fd = -1;
+
+      fd = open (rootfs, flags, mode);
+      if (UNLIKELY (fd < 0))
+        return crun_make_error (err, errno, "open `%s`", rootfs);
+
+      ret = check_fd_is_path (rootfs, fd, path, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
+      ret = fd;
+      fd = -1;
+      return ret;
+    }
+
   if (openat2_supported)
     {
     repeat:
@@ -449,7 +464,11 @@ crun_safe_ensure_at (bool do_open, bool dir, int dirfd, const char *dirpath,
 
   /* Empty path, nothing to do.  */
   if (*path == '\0')
-    return 0;
+    {
+      if (do_open)
+        return open (dirpath, O_CLOEXEC | O_PATH, 0);
+      return 0;
+    }
 
   npath = xstrdup (path);
 
@@ -537,7 +556,7 @@ crun_safe_ensure_at (bool do_open, bool dir, int dirfd, const char *dirpath,
               int saved_errno = errno;
 
               close (cwd);
-              return crun_make_error (err, saved_errno, "error stat'ing file `%s`", npath);
+              return crun_make_error (err, saved_errno, "stat `%s`", npath);
             }
           if ((st_mode & S_IFMT) != S_IFDIR)
             {
@@ -577,12 +596,12 @@ crun_safe_ensure_at (bool do_open, bool dir, int dirfd, const char *dirpath,
 int
 crun_safe_create_and_open_ref_at (bool dir, int dirfd, const char *dirpath, const char *path, int mode, libcrun_error_t *err)
 {
-  int fd;
+  int ret;
 
   /* If the file/dir already exists, just open it.  */
-  fd = safe_openat (dirfd, dirpath, path, O_PATH | O_CLOEXEC, 0, err);
-  if (LIKELY (fd >= 0))
-    return fd;
+  ret = safe_openat (dirfd, dirpath, path, O_PATH | O_CLOEXEC, 0, err);
+  if (LIKELY (ret >= 0))
+    return ret;
 
   crun_error_release (err);
   return crun_safe_ensure_at (true, dir, dirfd, dirpath, path, mode, MAX_READLINKS, err);
@@ -644,9 +663,9 @@ crun_dir_p_at (int dirfd, const char *path, bool nofollow, libcrun_error_t *err)
 
   ret = get_file_type_at (dirfd, &mode, nofollow, path);
   if (UNLIKELY (ret < 0))
-    return crun_make_error (err, errno, "error stat'ing file `%s`", path);
+    return crun_make_error (err, errno, "stat `%s`", path);
 
-  return S_ISDIR (mode);
+  return S_ISDIR (mode) ? 1 : 0;
 }
 
 int
@@ -702,7 +721,7 @@ static int selinux_enabled = -1;
 static int apparmor_enabled = -1;
 
 int
-libcrun_initialize_selinux (libcrun_error_t *err)
+libcrun_initialize_selinux (libcrun_container_t *container, libcrun_error_t *err)
 {
   cleanup_free char *out = NULL;
   cleanup_close int fd = -1;
@@ -712,11 +731,11 @@ libcrun_initialize_selinux (libcrun_error_t *err)
   if (selinux_enabled >= 0)
     return selinux_enabled;
 
-  fd = open ("/proc/mounts", O_RDONLY | O_CLOEXEC);
+  fd = libcrun_open_proc_file (container, "mounts", O_RDONLY, err);
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "open `/proc/mounts`");
+    return fd;
 
-  ret = read_all_fd_with_size_hint (fd, "/proc/mounts", &out, &len, get_page_size (), err);
+  ret = read_all_fd_with_size_hint (fd, "mounts", &out, &len, get_page_size (), err);
   if (UNLIKELY (ret < 0))
     return ret;
 
@@ -779,18 +798,15 @@ add_selinux_mount_label (char **retlabel, const char *data, const char *label, c
 }
 
 static const char *
-lsm_attr_path (const char *lsm, const char *fname, libcrun_error_t *err)
+lsm_attr_path (libcrun_container_t *container, const char *lsm, const char *fname, libcrun_error_t *err)
 {
   cleanup_close int attr_dirfd = -1;
   cleanup_close int lsm_dirfd = -1;
   char *attr_path = NULL;
 
-  attr_dirfd = open ("/proc/thread-self/attr", O_DIRECTORY | O_PATH | O_CLOEXEC);
+  attr_dirfd = libcrun_open_proc_file (container, "thread-self/attr", O_DIRECTORY | O_PATH, err);
   if (UNLIKELY (attr_dirfd < 0))
-    {
-      crun_make_error (err, errno, "open `/proc/thread-self/attr`");
-      return NULL;
-    }
+    return NULL;
 
   // Check for newer scoped interface in /proc/thread-self/attr/<lsm>
   if (lsm != NULL)
@@ -804,7 +820,7 @@ lsm_attr_path (const char *lsm, const char *fname, libcrun_error_t *err)
         }
     }
   // Use scoped interface if available, fall back to unscoped
-  xasprintf (&attr_path, "/proc/thread-self/attr/%s%s%s", lsm_dirfd >= 0 ? lsm : "", lsm_dirfd >= 0 ? "/" : "", fname);
+  xasprintf (&attr_path, "thread-self/attr/%s%s%s", lsm_dirfd >= 0 ? lsm : "", lsm_dirfd >= 0 ? "/" : "", fname);
 
   return attr_path;
 }
@@ -825,20 +841,19 @@ check_proc_super_magic (int fd, const char *path, libcrun_error_t *err)
 }
 
 static int
-set_security_attr (const char *lsm, const char *fname, const char *data, libcrun_error_t *err)
+set_security_attr (libcrun_container_t *container, const char *lsm, const char *fname, const char *data, libcrun_error_t *err)
 {
   int ret;
 
-  cleanup_free const char *attr_path = lsm_attr_path (lsm, fname, err);
+  cleanup_free const char *attr_path = lsm_attr_path (container, lsm, fname, err);
   cleanup_close int fd = -1;
 
   if (UNLIKELY (attr_path == NULL))
     return -1;
 
-  fd = open (attr_path, O_WRONLY | O_CLOEXEC);
-
+  fd = libcrun_open_proc_file (container, attr_path, O_WRONLY | O_CLOEXEC, err);
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "open `%s`", attr_path);
+    return fd;
 
   ret = check_proc_super_magic (fd, attr_path, err);
   if (UNLIKELY (ret < 0))
@@ -847,13 +862,13 @@ set_security_attr (const char *lsm, const char *fname, const char *data, libcrun
   // Write out data
   ret = TEMP_FAILURE_RETRY (write (fd, data, strlen (data)));
   if (UNLIKELY (ret < 0))
-    return crun_make_error (err, errno, "write file `%s`", attr_path);
+    return crun_make_error (err, errno, "write to file `%s`", attr_path);
 
   return 0;
 }
 
 int
-set_selinux_label (const char *label, bool now, libcrun_error_t *err)
+set_selinux_label (libcrun_container_t *container, const char *label, bool now, libcrun_error_t *err)
 {
   int ret;
 
@@ -862,7 +877,7 @@ set_selinux_label (const char *label, bool now, libcrun_error_t *err)
     return ret;
 
   if (ret)
-    return set_security_attr (NULL, now ? "current" : "exec", label, err);
+    return set_security_attr (container, NULL, now ? "current" : "exec", label, err);
   return 0;
 }
 
@@ -875,26 +890,25 @@ libcrun_is_apparmor_enabled (libcrun_error_t *err)
 }
 
 static int
-is_current_process_confined (libcrun_error_t *err)
+is_current_process_confined (libcrun_container_t *container, libcrun_error_t *err)
 {
-  cleanup_free const char *attr_path = lsm_attr_path ("apparmor", "current", err);
+  cleanup_free const char *attr_path = lsm_attr_path (container, "apparmor", "current", err);
   cleanup_close int fd = -1;
   char buf[256];
 
   if (UNLIKELY (attr_path == NULL))
     return -1;
 
-  fd = open (attr_path, O_RDONLY | O_CLOEXEC);
-
+  fd = libcrun_open_proc_file (container, attr_path, O_RDONLY | O_CLOEXEC, err);
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "open `%s`", attr_path);
+    return fd;
 
   if (UNLIKELY (check_proc_super_magic (fd, attr_path, err)))
     return -1;
 
   ssize_t bytes_read = read (fd, buf, sizeof (buf) - 1);
   if (UNLIKELY (bytes_read < 0))
-    return crun_make_error (err, errno, "error reading file `%s`", attr_path);
+    return crun_make_error (err, errno, "read from `%s`", attr_path);
 
 #define UNCONFINED "unconfined"
 #define UNCONFINED_LEN (ssize_t) (sizeof (UNCONFINED) - 1)
@@ -902,7 +916,7 @@ is_current_process_confined (libcrun_error_t *err)
 }
 
 int
-set_apparmor_profile (const char *profile, bool no_new_privileges, bool now, libcrun_error_t *err)
+set_apparmor_profile (libcrun_container_t *container, const char *profile, bool no_new_privileges, bool now, libcrun_error_t *err)
 {
   int ret;
 
@@ -913,7 +927,7 @@ set_apparmor_profile (const char *profile, bool no_new_privileges, bool now, lib
   if (ret)
     {
       cleanup_free char *buf = NULL;
-      ret = is_current_process_confined (err);
+      ret = is_current_process_confined (container, err);
       if (UNLIKELY (ret < 0))
         return ret;
       // if confined only way for apparmor to allow change of profile with NNP is with stacking
@@ -921,7 +935,7 @@ set_apparmor_profile (const char *profile, bool no_new_privileges, bool now, lib
                                                                          : "exec",
                  profile);
 
-      return set_security_attr ("apparmor", now ? "current" : "exec", buf, err);
+      return set_security_attr (container, "apparmor", now ? "current" : "exec", buf, err);
     }
   return 0;
 }
@@ -941,7 +955,7 @@ read_all_fd_with_size_hint (int fd, const char *description, char **out, size_t
     {
       ret = get_file_size (fd, &size);
       if (UNLIKELY (ret < 0))
-        return crun_make_error (err, errno, "error stat'ing file `%s`", description);
+        return crun_make_error (err, errno, "stat `%s`", description);
 
       allocated = size == 0 ? 1023 : size;
     }
@@ -953,7 +967,7 @@ read_all_fd_with_size_hint (int fd, const char *description, char **out, size_t
     {
       ret = TEMP_FAILURE_RETRY (read (fd, buf + nread, allocated - nread));
       if (UNLIKELY (ret < 0))
-        return crun_make_error (err, errno, "error reading from file `%s`", description);
+        return crun_make_error (err, errno, "read from file `%s`", description);
 
       if (ret == 0)
         break;
@@ -996,7 +1010,7 @@ read_all_file_at (int dirfd, const char *path, char **out, size_t *len, libcrun_
 
   fd = TEMP_FAILURE_RETRY (openat (dirfd, path, O_RDONLY | O_CLOEXEC));
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "error opening file `%s`", path);
+    return crun_make_error (err, errno, "open `%s`", path);
 
   return read_all_fd (fd, path, out, len, err);
 }
@@ -1017,7 +1031,7 @@ get_realpath_to_file (int dirfd, const char *path_name, char **absolute_path, li
 
   targetfd = TEMP_FAILURE_RETRY (openat (dirfd, path_name, O_RDONLY | O_CLOEXEC));
   if (UNLIKELY (targetfd < 0))
-    return crun_make_error (err, errno, "error opening file `%s`", path_name);
+    return crun_make_error (err, errno, "open `%s`", path_name);
   else
     {
       ssize_t len;
@@ -1048,20 +1062,24 @@ open_unix_domain_client_socket (const char *path, int dgram, libcrun_error_t *er
 
   fd = socket (AF_UNIX, dgram ? SOCK_DGRAM : SOCK_STREAM, 0);
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "error creating UNIX socket");
+    return crun_make_error (err, errno, "create UNIX socket");
 
   if (strlen (path) >= sizeof (addr.sun_path))
     {
       destfd = open (path, O_PATH | O_CLOEXEC);
       if (UNLIKELY (destfd < 0))
-        return crun_make_error (err, errno, "error opening `%s`", path);
+        return crun_make_error (err, errno, "open `%s`", path);
 
       get_proc_self_fd_path (name_buf, destfd);
 
       path = name_buf;
     }
 
-  strcpy (addr.sun_path, path);
+  size_t path_len = strlen (path);
+  if (path_len >= sizeof (addr.sun_path))
+    return crun_make_error (err, ENAMETOOLONG, "socket path too long: `%s`", path);
+
+  memcpy (addr.sun_path, path, path_len + 1);
   addr.sun_family = AF_UNIX;
   ret = connect (fd, (struct sockaddr *) &addr, sizeof (addr));
   if (UNLIKELY (ret < 0))
@@ -1081,14 +1099,19 @@ open_unix_domain_socket (const char *path, int dgram, libcrun_error_t *err)
   int ret;
   cleanup_close int fd = socket (AF_UNIX, dgram ? SOCK_DGRAM : SOCK_STREAM, 0);
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "error creating UNIX socket");
+    return crun_make_error (err, errno, "create UNIX socket");
 
   if (strlen (path) >= sizeof (addr.sun_path))
     {
       get_proc_self_fd_path (name_buf, fd);
       path = name_buf;
     }
-  strcpy (addr.sun_path, path);
+
+  size_t path_len = strlen (path);
+  if (path_len >= sizeof (addr.sun_path))
+    return crun_make_error (err, ENAMETOOLONG, "socket path too long: `%s`", path);
+
+  memcpy (addr.sun_path, path, path_len + 1);
   addr.sun_family = AF_UNIX;
   ret = bind (fd, (struct sockaddr *) &addr, sizeof (addr));
   if (UNLIKELY (ret < 0))
@@ -1500,7 +1523,7 @@ getsubidrange (uid_t id, int is_uid, uint32_t *from, uint32_t *len)
         }
 
       if (ret != ERANGE)
-        return ret;
+        return -1;
 
       buf_size *= 2;
       buf = xrealloc (buf, buf_size);
@@ -1538,37 +1561,55 @@ getsubidrange (uid_t id, int is_uid, uint32_t *from, uint32_t *len)
 
 #define MIN(x, y) ((x) < (y) ? (x) : (y))
 
-size_t
-format_default_id_mapping (char **ret, uid_t container_id, uid_t host_uid, uid_t host_id, int is_uid)
+int
+format_default_id_mapping (char **out, uid_t container_id, uid_t host_uid, uid_t host_id, int is_uid, libcrun_error_t *err)
 {
   uint32_t from = 0, available = 0;
   cleanup_free char *buffer = NULL;
-  size_t written = 0;
+  int written = 0;
+  int ret, remaining;
 
-  *ret = NULL;
+  *out = NULL;
 
   if (getsubidrange (host_uid, is_uid, &from, &available) < 0)
     return 0;
 
   /* More than enough space for all the mappings.  */
-  buffer = xmalloc (15 * 5 * 3);
+  remaining = 15 * 5 * 3;
+  buffer = xmalloc (remaining + 1);
 
   if (container_id > 0)
     {
       uint32_t used = MIN (container_id, available);
-      written += sprintf (buffer + written, "%d %d %d\n", 0, from, used);
+      ret = snprintf (buffer + written, remaining, "%d %d %d\n", 0, from, used);
+      if (UNLIKELY (ret >= remaining))
+        return crun_make_error (err, 0, "internal error: allocated buffer too small");
+
+      written += written;
+      remaining -= ret;
+
       from += used;
       available -= used;
     }
 
   /* Host ID -> Container ID.  */
-  written += sprintf (buffer + written, "%d %d 1\n", container_id, host_id);
+  ret = snprintf (buffer + written, remaining, "%d %d 1\n", container_id, host_id);
+  if (UNLIKELY (ret >= remaining))
+    return crun_make_error (err, 0, "internal error: allocated buffer too small");
+  written += ret;
+  remaining -= ret;
 
   /* Last mapping: use any id that is left.  */
   if (available)
-    written += sprintf (buffer + written, "%d %d %d\n", container_id + 1, from, available);
+    {
+      ret = snprintf (buffer + written, remaining, "%d %d %d\n", container_id + 1, from, available);
+      if (UNLIKELY (ret >= remaining))
+        return crun_make_error (err, 0, "internal error: allocated buffer too small");
+      written += ret;
+      remaining -= ret;
+    }
 
-  *ret = buffer;
+  *out = buffer;
   buffer = NULL;
   return written;
 }
@@ -1594,7 +1635,7 @@ run_process_child (char *path, char **args, const char *cwd, char **envp, int pi
   int dev_null_fd = -1;
   int ret;
 
-  ret = mark_or_close_fds_ge_than (3, false, &err);
+  ret = mark_or_close_fds_ge_than (NULL, 3, false, &err);
   if (UNLIKELY (ret < 0))
     libcrun_fail_with_error ((err)->status, "%s", (err)->msg);
 
@@ -1686,7 +1727,7 @@ run_process_with_stdin_timeout_envp (char *path, char **args, const char *cwd, i
          been terminated.  */
       if (errno != EPIPE)
         {
-          ret = crun_make_error (err, errno, "writing to pipe");
+          ret = crun_make_error (err, errno, "write to pipe");
           goto restore_sig_mask_and_exit;
         }
     }
@@ -1750,7 +1791,7 @@ restore_sig_mask_and_exit:
 }
 
 int
-mark_or_close_fds_ge_than (int n, bool close_now, libcrun_error_t *err)
+mark_or_close_fds_ge_than (libcrun_container_t *container, int n, bool close_now, libcrun_error_t *err)
 {
   cleanup_close int cfd = -1;
   cleanup_dir DIR *dir = NULL;
@@ -1760,21 +1801,25 @@ mark_or_close_fds_ge_than (int n, bool close_now, libcrun_error_t *err)
 
   ret = syscall_close_range (n, UINT_MAX, close_now ? 0 : CLOSE_RANGE_CLOEXEC);
   if (ret == 0)
-    return 0;
+    {
+      if (close_now && container->proc_fd >= n)
+        container->proc_fd = -1;
+      return 0;
+    }
   if (ret < 0 && errno != EINVAL && errno != ENOSYS && errno != EPERM)
     return crun_make_error (err, errno, "close_range from `%d`", n);
 
-  cfd = open ("/proc/self/fd", O_DIRECTORY | O_RDONLY | O_CLOEXEC);
+  cfd = libcrun_open_proc_file (container, "self/fd", O_DIRECTORY | O_RDONLY, err);
   if (UNLIKELY (cfd < 0))
-    return crun_make_error (err, errno, "open `/proc/self/fd`");
+    return cfd;
 
-  ret = check_proc_super_magic (cfd, "/proc/self/fd", err);
+  ret = check_proc_super_magic (cfd, "self/fd", err);
   if (UNLIKELY (ret < 0))
     return ret;
 
   dir = fdopendir (cfd);
   if (UNLIKELY (dir == NULL))
-    return crun_make_error (err, errno, "cannot fdopendir `/proc/self/fd`");
+    return crun_make_error (err, errno, "fdopendir `self/fd`");
 
   /* Now it is owned by dir.  */
   cfd = -1;
@@ -1795,13 +1840,16 @@ mark_or_close_fds_ge_than (int n, bool close_now, libcrun_error_t *err)
         {
           ret = close (val);
           if (UNLIKELY (ret < 0))
-            return crun_make_error (err, errno, "close(fd=%d)", val);
+            return crun_make_error (err, errno, "close fd `%d`", val);
+
+          if (val == container->proc_fd)
+            container->proc_fd = -1;
         }
       else
         {
           ret = fcntl (val, F_SETFD, FD_CLOEXEC);
           if (UNLIKELY (ret < 0))
-            return crun_make_error (err, errno, "cannot set CLOEXEC fd for `/proc/self/fd/%s`", name);
+            return crun_make_error (err, errno, "cannot set CLOEXEC for fd `%d`", val);
         }
     }
   return 0;
@@ -1818,7 +1866,7 @@ get_current_timestamp (char *out, size_t len)
   gmtime_r (&tv.tv_sec, &now);
   strftime (timestamp, sizeof (timestamp), "%Y-%m-%dT%H:%M:%S", &now);
 
-  snprintf (out, len, "%s.%06lldZ", timestamp, (long long int) tv.tv_usec);
+  (void) snprintf (out, len, "%s.%06lldZ", timestamp, (long long int) tv.tv_usec);
   out[len - 1] = '\0';
 }
 
@@ -1990,7 +2038,7 @@ fail:
       errno = -last_error;
       if (errno == ENOENT)
         return crun_make_error (err, errno, "executable file `%s` not found%s", executable_path, executable_path[0] == '/' ? "" : " in $PATH");
-      return crun_make_error (err, errno, "cannot open `%s`", executable_path);
+      return crun_make_error (err, errno, "open `%s`", executable_path);
     }
 }
 
@@ -2044,7 +2092,7 @@ copy_xattr (int sfd, int dfd, const char *srcname, const char *destname, libcrun
       if (errno == ENOTSUP)
         return 0;
 
-      return crun_make_error (err, errno, "get xattr list for `%s`", srcname);
+      return crun_make_error (err, errno, "flistxattr `%s`", srcname);
     }
 
   if (xattr_len == 0)
@@ -2054,7 +2102,7 @@ copy_xattr (int sfd, int dfd, const char *srcname, const char *destname, libcrun
 
   xattr_len = flistxattr (sfd, buf, xattr_len + 1);
   if (UNLIKELY (xattr_len < 0))
-    return crun_make_error (err, errno, "get xattr list for `%s`", srcname);
+    return crun_make_error (err, errno, "flistxattr `%s`", srcname);
 
   for (it = buf; it - buf < xattr_len; it += strlen (it) + 1)
     {
@@ -2071,7 +2119,7 @@ copy_xattr (int sfd, int dfd, const char *srcname, const char *destname, libcrun
           if (errno == EINVAL || errno == EOPNOTSUPP)
             continue;
 
-          return crun_make_error (err, errno, "set xattr for `%s`", destname);
+          return crun_make_error (err, errno, "fsetxattr `%s` to `%s`", it, destname);
         }
     }
 
@@ -2133,7 +2181,7 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
   if (UNLIKELY (dsrcfd == NULL))
     {
       TEMP_FAILURE_RETRY (close (srcdirfd));
-      return crun_make_error (err, errno, "cannot open directory `%s`", destname);
+      return crun_make_error (err, errno, "open directory `%s`", destname);
     }
 
   for (de = readdir (dsrcfd); de; de = readdir (dsrcfd))
@@ -2212,7 +2260,7 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
 
           ret = symlinkat (target_buf, destdirfd, de->d_name);
           if (UNLIKELY (ret < 0))
-            return crun_make_error (err, errno, "create symlink `%s/%s`", destname, de->d_name);
+            return crun_make_error (err, errno, "symlinkat `%s/%s`", destname, de->d_name);
           break;
 
         case S_IFBLK:
@@ -2221,13 +2269,13 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
         case S_IFSOCK:
           ret = mknodat (destdirfd, de->d_name, mode, rdev);
           if (UNLIKELY (ret < 0))
-            return crun_make_error (err, errno, "create special file `%s/%s`", destname, de->d_name);
+            return crun_make_error (err, errno, "mknodat `%s/%s`", destname, de->d_name);
           break;
         }
 
       ret = fchownat (destdirfd, de->d_name, uid, gid, AT_SYMLINK_NOFOLLOW);
       if (UNLIKELY (ret < 0))
-        return crun_make_error (err, errno, "chown `%s/%s`", destname, de->d_name);
+        return crun_make_error (err, errno, "fchownat `%s/%s`", destname, de->d_name);
 
       /*
        * ALLPERMS is not defined by POSIX
@@ -2243,7 +2291,7 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
           if (errno == ENOTSUP)
             continue;
 
-          return crun_make_error (err, errno, "chmod `%s/%s`", destname, de->d_name);
+          return crun_make_error (err, errno, "fchmodat `%s/%s`", destname, de->d_name);
         }
     }
 
@@ -2270,7 +2318,7 @@ safe_write (int fd, const char *fname, const void *buf, size_t count, libcrun_er
         {
           if (errno == EINTR || errno == EAGAIN)
             continue;
-          return crun_make_error (err, errno, "write file `%s`", fname);
+          return crun_make_error (err, errno, "write to `%s`", fname);
         }
       written += w;
     }
@@ -2357,11 +2405,17 @@ append_paths (char **out, libcrun_error_t *err, ...)
   return 0;
 }
 
+#if __has_attribute(__nonstring__)
+#  define __nonstring __attribute__ ((__nonstring__))
+#else
+#  define __nonstring
+#endif
+
 /* Adapted from mailutils 0.6.91 (distributed under LGPL 2.0+)  */
 static int
 b64_input (char c)
 {
-  const char table[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
+  const char table[64] __nonstring = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
   int i;
 
   for (i = 0; i < 64; i++)
@@ -2440,10 +2494,13 @@ has_suffix (const char *str, const char *suffix)
 char *
 str_join_array (int offset, size_t size, char *const array[], const char *joint)
 {
-  size_t jlen, lens[size];
+  size_t jlen;
+  cleanup_free size_t *lens = NULL;
   size_t i, total_size = (size - 1) * (jlen = strlen (joint)) + 1;
   char *result, *p;
 
+  lens = xmalloc (size * sizeof (*lens));
+
   for (i = 0; i < size; ++i)
     {
       total_size += (lens[i] = strlen (array[i]));
@@ -2711,7 +2768,7 @@ channel_fd_pair_process (struct channel_fd_pair *channel, int epollfd, libcrun_e
   for (i = 0, repeat = true; i < 1000 && repeat; i++)
     {
       repeat = false;
-      if (ring_buffer_get_space_available (channel->rb) >= ring_buffer_get_size (channel->rb))
+      if (ring_buffer_get_space_available (channel->rb) > 0)
         {
           ret = ring_buffer_read (channel->rb, channel->in_fd, &is_input_eagain, err);
           if (UNLIKELY (ret < 0))
@@ -2758,3 +2815,56 @@ channel_fd_pair_process (struct channel_fd_pair *channel, int epollfd, libcrun_e
     }
   return 0;
 }
+
+int
+libcrun_get_cached_proc_fd (libcrun_container_t *container, libcrun_error_t *err)
+{
+  if (container->proc_fd < 0)
+    {
+#ifdef HAVE_NEW_MOUNT_API
+      cleanup_close int fsfd = -1;
+      int ret;
+
+      fsfd = syscall_fsopen ("proc", FSOPEN_CLOEXEC);
+      if (fsfd >= 0)
+        {
+          ret = syscall_fsconfig (fsfd, FSCONFIG_CMD_CREATE, NULL, NULL, 0);
+          if (ret >= 0)
+            {
+              container->proc_fd = syscall_fsmount (fsfd, FSMOUNT_CLOEXEC, 0);
+              if (container->proc_fd >= 0)
+                return container->proc_fd;
+            }
+        }
+#endif
+
+      container->proc_fd = open ("/proc", O_RDONLY | O_DIRECTORY | O_CLOEXEC);
+      if (container->proc_fd < 0)
+        return crun_make_error (err, errno, "open `/proc`");
+    }
+  return container->proc_fd;
+}
+
+int
+libcrun_open_proc_file (libcrun_container_t *container, const char *path, int flags, libcrun_error_t *err)
+{
+  int proc_fd, fd;
+
+  proc_fd = libcrun_get_cached_proc_fd (container, err);
+  if (proc_fd < 0)
+    return proc_fd;
+
+  fd = TEMP_FAILURE_RETRY (openat (proc_fd, path, flags | O_CLOEXEC));
+  if (UNLIKELY (fd < 0))
+    return crun_make_error (err, errno, "openat `/proc/%s`", path);
+
+  return fd;
+}
+
+int
+libcrun_open_proc_pid_file (libcrun_container_t *container, pid_t pid, const char *path, int flags, libcrun_error_t *err)
+{
+  cleanup_free char *full_path = NULL;
+  xasprintf (&full_path, "%d/%s", pid, path);
+  return libcrun_open_proc_file (container, full_path, flags, err);
+}
